/* The shadow of the original with only common prototypes now. */
#include <stdio.h>
#include <sys/types.h>

/* get definition of HZ */
#include <asm/param.h>

/* get page info */
#include <asm/page.h>

extern char *wchan(unsigned long);
extern char *find_func(unsigned long address);
extern void *xcalloc(void *pointer, int size);
extern void *xmalloc(unsigned int size);
extern void *xrealloc(void *oldp, unsigned int size);

extern int mult_lvl_cmp(void *a, void *b);
extern int node_mult_lvl_cmp(void *a, void *b);
extern void dump_keys(void);

extern char *user_from_uid(int uid);

extern int open_sysmap(void);
extern int open_psdb(void);
extern void close_psdb(void);
extern void make_fnctbl(void);

extern unsigned print_str(FILE * file, char *s, unsigned max);
extern unsigned print_strlist(FILE * file, char **strs, char *sep, unsigned max);
extern unsigned snprint_strlist(char *buf, int max, char **strs, char *sep);
